"""
A module for computational geometry.
"""
from shapes import *
from standalone import *
from locators import *
from rtree import *
from kdtree import *
from sphere import *
#why don't we import collection?
